#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Mar 23 10:11:35 2023

@author: Izotov Yuriy
@user: izotov93
"""

import logging

logging.basicConfig(
    level=logging.INFO,
    format='[%(asctime)s]:%(levelname)s: %(message)s')

import numpy as np
from numba import njit
from sys import exit

def read_mnist_images(file_name : str, limit=1):
    """
    Reading their MNIST images
        :param file_name: Image file name mnist. Format idx3-ubyte
        :param limit: Usage fraction of the selected database
        :return: Linear array containing image (0 to 255)
    """
    intType = np.dtype('int32').newbyteorder('>')
    nMetaDataBytes = 4 * intType.itemsize
    images = np.fromfile(file_name, dtype='ubyte')
    magicBytes, nImages, width, height = \
        np.frombuffer(images[:nMetaDataBytes].tobytes(), intType)
    images = images[nMetaDataBytes:].astype(
        dtype='uint8').reshape([nImages, width, height])
    # Restriction on reading data
    limit = round(limit * images.shape[0])
    images = images[:limit]

    # Rotating and Mirroring an Image
    images = [np.flipud(np.rot90(i[:], 3, axes=(1, 0)))
                .reshape(width * height) for i in images]

    return np.array(images)

def read_mnist_labels(file_name : str, limit=1):
    """
    Reading their MNIST labels
        :param file_name: Labels file name mnist. Format idx1-ubyte
        :param limit: Usage fraction of the selected database
        :return: Linear array containing labels
    """
    intType = np.dtype('int32').newbyteorder('>')
    labels = np.fromfile(file_name, dtype='ubyte')[2 * intType.itemsize:]
    # Restriction on reading data
    limit = round(limit * labels.shape[0])
    labels = labels[:limit]

    return labels

def T_pattern_load(pattern_file : str):
    """
    Loading a Pattern
        :param pattern_file: Pattern file name mnist. Format .VM
        :return: A linear array containing a pattern of indexes, for converting images to mnist
    """
    logging.debug('Loading pattern file %s', pattern_file)
    try:
        with open(pattern_file) as file:
            sire_ar = int(file.readline())
            data_list = []
            for line in file:
                st_list = line.replace('\n', '').split('\t')
                data_list.append(st_list)
    except:
        logging.info('Error read file %s', pattern_file)
        exit(0)
    pattern = []
    for el in data_list:
        pattern.append((int(el[1]) - 1) * 28 + int(el[0]))

    return np.array(pattern, dtype='int')

@njit(cache=True, fastmath=True)
def T_pattern_application(data : np.ndarray, pattern : np.ndarray):
    """
    Applying a pattern on one element mnist
        :param data: Array containing image mnist
        :param pattern: Array containing pattern
        :return: Image MNIST converted
    """
    out_data = np.empty_like(data)
    for i in range(data.size):
        out_data[i] = data[pattern[i]-1]
    return out_data

def data_pattern_transform(database, pattern_file :str):
    """
    The function of applying a pattern to the entire image database
        :param database: database containing image mnist
        :param pattern_file: Path to the pattern file
        :return: Prepared database MNIST (0 to 255) + bias
    """
    pattern = T_pattern_load(pattern_file)
    logging.debug('Pattern - %s application', pattern_file)
    database_pattern = np.empty_like(database)
    database_pattern = np.insert(database_pattern, 0, 1, axis=1)
    for index, data in enumerate(database):
        database_pattern[index] = np.insert(
            T_pattern_application(data, pattern), 0, 1)

    return database_pattern

def read_database_covid19(file_name: str, limit=1):
    """
    Covid-19 database reading function
        :param file_name: Path to the SARS-CoV-2-RBV1 file
        :param limit: usage fraction of the selected dataset
        :return: train_data and test_data: Array with data and labels;
                 max_data and min_data: Array of maximum and minimum database values along axis 0.
    """
    try:
        train_data = np.loadtxt(file_name, delimiter='\t')
    except:
        logging.error('File %s database not found', file_name)
        exit(0)

    # Restriction on reading data
    limit = round(limit * train_data.shape[0])
    train_data = train_data[:limit]

    test_data = train_data[:, -1]
    test_data = np.stack([np.zeros(test_data.__len__()),
                          test_data]).T
    for data in test_data:
        if (data[-1] != 1):
            data[0] = 1

    train_data = np.delete(train_data, -1, 1)
    train_data = np.insert(train_data, 0, 1, axis=1)

    max_data = np.zeros(train_data.shape[1])  # Y
    min_data = np.zeros(train_data.shape[1])  # Y
    min_data[:] = 1000

    for data in train_data:
        max_data = np.maximum(data, max_data)
        min_data = np.minimum(data, min_data)

    return train_data, test_data, max_data, min_data

def read_database_mnist(file_name_image: str,
                        file_name_labels: str,
                        file_pattern : str,
                        limit=1):
    """
    MNIST database reading function
        :param file_name_image: Path to the image mnist
        :param file_name_labels: Path to the label mnist
        :param file_pattern: Path to the pattern
        :param limit: usage fraction of the selected dataset
        :return: train_data, test_data and test_data_array: Array with data and labels;
                max_data and min_data: Array of maximum and minimum database values along axis 0.
    """
    try:
        train_data = read_mnist_images(file_name_image,
                                       limit)
        test_data = read_mnist_labels(file_name_labels,
                                      limit)
    except:
        logging.error('Error read MNIST database')
        exit(0)
    train_data = data_pattern_transform(train_data, file_pattern)

    test_data_array = np.zeros((test_data.shape[0], 10))
    for index in range(test_data.shape[0]):
        test_data_array[index][test_data[index]] = 1

    max_data = np.zeros(train_data.shape[1])  # Y
    min_data = np.zeros(train_data.shape[1])  # Y
    min_data[:] = 1000
    for data in train_data:
        max_data = np.maximum(data, max_data)
        min_data = np.minimum(data, min_data)

    return train_data, test_data, test_data_array, max_data, min_data

@njit(cache=True, fastmath=True)
def normalization(input, max_data, min_data):
    """
    Single element normalization
        :param input: Database element
        :param max_data: Array of maximum database values
        :param min_data: Array of minimal database values
        :return: Normalized array (0 to 1)
    """
    norm_data = input.copy()
    norm_data[0] = 1
    for i in range(1, input.size):
        if ((max_data[i] - min_data[i]) != 0):
            norm_data[i] = (input[i] - min_data[i]) / \
                        (max_data[i] - min_data[i])
        elif (max_data[i] != 0):
            norm_data[i] = (input[i] - min_data[i]) / max_data[i]
        else:
            norm_data[i] = 0
    return norm_data

# Step 2
@njit(cache=True, fastmath=True)
def format_norm_database(database, max_data, min_data):
    """
    Database normalization function
        :param database: Array containing database elements
        :param max_data: Array of maximum database values
        :param min_data: Array of minimal database values
        :return: Normalized database (0 to 1)
    """
    norm_database = np.empty_like(database, dtype=np.float64)
    for index, data in enumerate(database):
        norm_database[index] = normalization(np.float64(data),
                                             np.float64(max_data),
                                             np.float64(min_data))
    return norm_database

def main():
    pass

if __name__ == '__main__':
    main()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Mar 23 10:11:35 2023

@author: Izotov Yuriy
@user: izotov93
"""

import numpy as np
from numba import njit, types
import math

@njit(cache=True, fastmath=True)
def Fout(input : types.float64[:]) -> types.float64[:]:
    """
    Logistic activation function (1 / (1 + exp(-1 * value))
        :param input: The meaning of the neuron
        :return: Neuron value after activation
    """
    return np.divide(1, np.add(1, np.exp(-input)))

@njit(cache=True, fastmath=True)
def back_prop_calc_first_layer(Y, W, Sh_max,
                               Sh_min, Sh_mean):
    """
    Calculation of neurons of the first hidden layer
        :param Y: Input vector
        :param W: Weight Matrix
        :param Sh_max, Sh_min and Sh_mean: Normalization coefficient
        :return: The value of the neurons of one hidden layer
    """
    Sh = np.sum(np.multiply(Y, W), axis=1)
    Sh = np.divide(Sh - Sh_min, Sh_max - Sh_min) \
                    - 0.5 - Sh_mean
    return np.hstack((np.ones(1), Sh)) # bias

@njit(cache=True, fastmath=True)
def back_prop_calc_value(S, W):
    """
    Calculation of neurons in hidden layers (second and further)
        :param S: The value of the neurons of the past hidden layer
        :param W: Weight Matrix
        :return: The value of the neurons of hidden layer
    """
    Sh_next = np.sum(np.multiply(S, W), axis=1)
    return Fout(Sh_next)

@njit(cache=True, fastmath=True)
def weight_training(W, err, Sh, NORMA):
    """
    Weight training function
        :param W: Weight Matrix
        :param err: Error vector computed backpropagation method
        :param Sh: The value of the neurons of hidden layer
        :param NORMA: -
        :return: Weight matrix after backpropagation
    """
    dW = np.empty_like(W)
    dW[:] = Sh * NORMA
    dW2 = dW * err.reshape((-1, 1))
    return np.add(W, dW2)

@njit(cache=True, fastmath=True)
def err_calc(Sh: types.float64[:]) -> types.float64[:]:
    """
    Calculate errors
        :param Sh: The value of the neurons of hidden layer
        :return: Error vector
    """
    return Sh * (1 - Sh)

@njit(cache=True, fastmath=True)
def err_calc_hid_layers(err: types.float64[:], W : types.float64[:]) \
        -> types.float64[:]:
    """
    Error Calculation on Hidden Layers (intermediate function)
        :param err: Error vector
        :param W: Weight Matrix
        :return: Error vector
    """
    return np.sum(np.multiply(err, W.T), axis=1)

@njit(cache=True, fastmath=True)
def err_calc_last_layer(techer, Sh):
    """
    Error Calculation on Last Layer
        :param techer: reference vector
        :param Sh: The value of the neurons of hidden layer
        :return: Error vector
    """
    return (techer - Sh) * err_calc(Sh)

@njit(cache=True, fastmath=True)
def err_calc_rest_layers(Sh, err_prev, W_prev):
    """
    Error Calculation on Hidden Layers
        :param Sh: The value of the neurons of hidden layer
        :param err_prev: The error vector of the previous layer
        :param W_prev: Weight matrix of the previous layer
        :return: Error vector
    """
    return err_calc(Sh) * err_calc_hid_layers(err_prev, W_prev)

# Step 3
@njit(cache=True, fastmath=True)
def formation_W1_coef_time_series(xn_params, W1, method=5):
    """
    Formation of the weight matrix from the time series
        :param xn_params: Time series
        :param W1: Weight matrix
        :param method: One of 6 methods for forming a reservoir matrix
        :return: Weight matrix W1
    """
    Y = W1.shape[1]#conf_layers[0] + 1
    P = W1.shape[0]#conf_layers[1]
    k = 0
    if (method == 1): # line by line doubling
        for i in range(P):
            j = 0
            while (j < Y):
                if (k < xn_params.size):
                    W1[i, j] = xn_params[k]
                else:
                    k = -1
                    j -= 1
                k += 1
                j += 1
    elif (method == 2): # Line by line with zeros
        for i in range(P):
            j = 0
            while (j < Y):
                if (k < xn_params.size):
                    W1[i, j] = xn_params[k]
                else:
                    k = -1
                    j = Y
                k += 1
                j += 1
            if (k >= xn_params.size):
                k = 0
    elif (method == 3): # Line Stretch
        XXk = np.empty_like(xn_params)
        for index in range(XXk.size):
            XXk[index] = ((index/(XXk.size-1)) *
                          ((Y*P)-1))+1
        for i in range(P):
            for j in range(Y):
                z = 0
                k += 1
                while ((z < xn_params.size-1) and
                       (int(XXk[z]) <= k)):
                    z += 1
                if z == 0:
                    z = 1
                x1 = XXk[z - 1]
                x2 = XXk[z]
                y1 = xn_params[z - 1]
                y2 = xn_params[z]
                if x1 == x2:
                    koef = 0
                else:
                    koef = (y1-y2)/(x1-x2)
                b_kof = y1 - koef * x1
                W1[i, j] = koef * k + b_kof
    elif (method == 4):
        for i in range(Y):
            j = 0
            while (j < P):
                if (k < xn_params.size):
                    W1[j, i] = xn_params[k]
                else:
                    k = -1
                    j -= 1
                j += 1
                k += 1
    elif (method == 5):
        for i in range(Y):
            j = 0
            while (j < P):
                if (k < xn_params.size):
                    W1[j, i] = xn_params[k]
                else:
                    j = P
                j += 1
                k += 1
            if (k >= xn_params.size):
                k = 0
    elif (method == 6):
        XXk = np.empty_like(xn_params)
        for index in range(XXk.size):
            XXk[index] = ((index / (XXk.size - 1)) *
                          ((Y * P) - 1)) + 1
        for i in range(Y):
            for j in range(P):
                k += 1
                z = 0
                while ((z < xn_params.size-1) and
                       (int(XXk[z]) <= k)):
                    z += 1
                if z == 0:
                    z = 1
                x1 = XXk[z - 1]
                x2 = XXk[z]
                y1 = xn_params[z - 1]
                y2 = xn_params[z]
                if x1 == x2:
                    koef = 0
                else:
                    koef = (y1-y2)/(x1-x2)
                b_kof = y1 - koef * x1
                W1[j, i] = koef * k + b_kof

    return W1

# Step 4
@njit(cache=True, fastmath=True)
def calculation_min_max_params(database, W1):
    """
    Calculation of minimum and maximum normalization vectors
        :param database: database
        :param W1: Weight matrix W1
        :return: minimum and maximum normalization vectors
    """
    Sh_max = np.zeros(W1.shape[0], dtype=np.float64)  # P
    Sh_min = np.empty(W1.shape[0], dtype=np.float64)  # P
    Sh_min[:] = 100
    for data in database:
        # We multiply element by element Y by the transposed matrix W
        # We get the matrix and find the sum along axis 1
        Sh = np.sum(np.multiply(data, W1), axis=1)

        Sh_max = np.maximum(Sh, Sh_max)
        Sh_min = np.minimum(Sh, Sh_min)
    return Sh_max, Sh_min

#Step 5
#@njit(cache=True, fastmath=True)
def calculation_normalization_params(database, W1):
    """
    Calculation normalization params
        :param database: Normalized database
        :param W1: Weight matrix W1
        :return: minimum, maximum and mean normalization vectors
    """
    Sh_max, Sh_min = calculation_min_max_params(database, W1)
    # If the base is less than 1000, then the entire base
    meanS_size = min([1000, database.shape[0]])
    meanS = np.zeros((meanS_size, W1.shape[0]), dtype=np.float64)

    for index in range(meanS.shape[0]):
        meanS[index] = np.sum(np.multiply(database[index], W1), axis=1)
        div = np.divide((meanS[index] - Sh_min), (Sh_max - Sh_min),
                        where=(Sh_max-Sh_min)!=0)
        meanS[index] = np.subtract(div, 0.5, where=div!=0)

    return Sh_max, Sh_min, np.mean(meanS, axis=0)

# Step 6
def formation_matrix_Wn(num, conf_layers, xr_coef=0):
    a = conf_layers[num - 1] + 1
    b = conf_layers[num]
    if (conf_layers.__len__() - 1 - num) != 0:
        b += 1

    W_list = np.empty(a * b, dtype=np.float64)
    if (xr_coef != 0):
        for i in range(W_list.__len__()):
            xr_coef = 1 - 1.95 * np.square(xr_coef)
            W_list[i] = xr_coef
    else:
        W_list[:] = 0.5
    W_list = np.reshape(W_list, (b, a))

    return W_list

def formation_reservoir_coef(conf_layers, LCG):
    """
    Formation of the matrix of weights using the linear congruential generator
        :param conf_layers: Neural network layer configuration
        :param LCG: List to coefficients the linear congruential generator
        :return: Weight matrix W1
    """
    W1 = np.empty((conf_layers[0] + 1, conf_layers[1]))  # (Y+1, P)
    # LCG = ("K", "D", "L", "C")
    zn = LCG[3]
    for j in range(conf_layers[1]): # P
        for i in range(conf_layers[0] + 1): # Y
            # W = (D - K * W) % L;
            zn = int(math.fmod((LCG[1] - LCG[0] * zn), LCG[2]))
            W1[i, j] = zn / LCG[2]
    return W1.T

# Step 8
@njit(cache=True, fastmath=True)
def calculation_neuron_first_layer(norm_data, W1,
                                   Sh_max, Sh_min, Sh_mean):
    """
    Calculation of neurons of the first hidden layer over the entire database
        :param norm_data: Normalized database
        :param W1: Weight matrix W1
        :param Sh_max, Sh_min and Sh_mean: Normalization coefficient
        :return: Matrix Sh containing the values of the neurons of the first hidden layer
    """
    Sh = np.zeros((norm_data.shape[0], Sh_max.shape[0]+1),
                  dtype=np.float64)
    for index in range(norm_data.shape[0]):
        Sh[index] = back_prop_calc_first_layer(norm_data[index], W1,
                                            Sh_max, Sh_min, Sh_mean)
    return Sh

@njit(cache=True, fastmath=True)
def calc_metric(model : np.ndarray, Sout: np.ndarray,
                metric : str) -> np.float64:
    """
    Calculating the entropy value from a metric
        :param model: Reference Vector
        :param Sout: The result of the neural network
        :param metric, 'Acc' - accuracy metric,
                    'R2E' - R2 Efficiency metric,
                    'PE' - Pearson Efficiency metric.
        :return: NNetEn – the entropy value
    """

    if metric == 'Acc':
        if (np.argmax(model) == np.argmax(Sout)):
            return 1
        else:
            return 0
    elif metric == 'R2E':
        R2 = 1 - (np.sum(np.square(model - Sout)) / \
                np.sum(np.square(model-np.mean(model))))
        return R2
    elif metric == 'PE':
        ro = np.sum((model - np.mean(model))*(Sout- np.mean(Sout))) / \
                (np.sqrt(np.sum(np.square(model-np.mean(model))))*
                    np.sqrt(np.sum(np.square(Sout - np.mean(Sout)))))
        return ro


def main():
    pass

if __name__ == '__main__':
    main()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Mar 23 10:11:35 2023

@author: Izotov Yuriy
@user: izotov93
"""

import logging

logging.basicConfig(
    level=logging.INFO,
    format='[%(asctime)s]:%(levelname)s: %(message)s')

from time import time
from .database import read_database_mnist, \
    read_database_covid19, format_norm_database
from .LogNNet import *
from numba.typed import List
import os
from sys import exit

class init_database(object):
    __dir = os.path.dirname(os.path.abspath(__file__))
    __path_database = os.path.join(__dir, 'Database')

    def __init__(self, database='D1', mu=1):
        """
        Class initialization with datasets
            :param database: (default = D1) Select dataset,
                            D1 – MNIST;
                            D2 – SARS-CoV-2-RBV1.
            :param mu: (default = 1) usage fraction of the selected dataset (0.01 .. 1).
            :return: Instance of class init_database
        """

        self.__validation(database=database, mu=mu)

        if database == 'D1':
            self.__read_mnist_database(mu=mu)
        elif database == 'D2':
            self.__read_covid19_database(mu=mu)

    def __validation(self, database : str, mu : float):
        """
        Validation of input data
            :param database: type -> str
            :param mu: type -> float
            :return: If the validation of the type and data was
                not successful, an error message will be displayed
        """
        str_db = {'D1' : 'MNIST',
                  'D2' : 'SARS-CoV-2-RBV1'}.get(database)
        if str_db:
            logging.debug('Loading database {}'
                     ' [Mu - {}]'.format(str_db, mu))
        else:
            logging.error('Invalid parameter - database')
            exit(0)

        if ((isinstance(mu, float) or isinstance(mu, int))
                and (mu >= 0.01 and mu <= 1)):
            self.mu = mu
        else:
            logging.error('Invalid parameter - mu')
            exit(0)

    def __read_mnist_database(self, mu : int):
        """
        Read database mnist and data normalization (0 to 1)
            :param mu: Usage fraction of the selected database (0.01 to 1)
            :return: Arrays of training and test data, as well as their labels.
        """
        # Step 1
        file_pattern = 'T-pattern-3.VM'

        # Train database
        dbase, test_db, self.lab_train, max_data, min_data = \
            read_database_mnist(os.path.join(self.__path_database,
                                         'train-images.idx3-ubyte'),
                                os.path.join(self.__path_database,
                                             'train-labels.idx1-ubyte'),
                                os.path.join(self.__path_database,
                                             file_pattern),
                                limit=mu)
        # Test database
        dbase_T, test_db_T, self.lab_test, max_data, min_data = \
            read_database_mnist(os.path.join(self.__path_database,
                                        't10k-images.idx3-ubyte'),
                                os.path.join(self.__path_database,
                                             't10k-labels.idx1-ubyte'),
                                os.path.join(self.__path_database,
                                             file_pattern),
                                limit=mu)
        # Step 2
        logging.debug('Step 2. Formation the normalization database')
        # Train labels
        # FIXME
        self.norm_db = dbase / 255
        self.norm_db[:, 0] = 1
        # norm_data = format_norm_database(dbase, max_data, min_data)

        # Test labels
        # FIXME
        self.norm_db_T = dbase_T / 255
        self.norm_db_T[:, 0] = 1
        # norm_data_T = format_norm_database(dbase_T, max_data, min_data)

    def __read_covid19_database(self, mu : int):
        """
        Read database SARS-CoV-2-RBV1 and data normalization (0 to 1)
            :param mu: Usage fraction of the selected database (0.01 to 1)
            :return: Arrays of training and test data, as well as their labels.
        """
        # Step 1
        dbase, self.lab_train, max_data, min_data = \
            read_database_covid19(os.path.join(self.__path_database,
                                        'SARS-CoV-2-RBV1.txt'),
                                  limit=mu)
        # Step 2
        logging.debug('Step 2. Formation the normalization database')
        self.norm_db = format_norm_database(dbase, max_data, min_data)

        self.norm_db_T, self.lab_test = \
            self.norm_db, self.lab_train

class NNetEn_entropy(init_database):
    # BASIC PARAMS
    __NORMA = 0.2
    __NEURON_HID_LAYER = 25

    def __init__(self, database='D1', mu=1):
        """
        Initialization of the database by input parameters, formation of the
        basic parameters of the neural network LogNNet
            :param database: (default = D1) Select dataset: D1 – MNIST, D2 – SARS-CoV-2-RBV1.
            :param mu: (default = 1) usage fraction of the selected dataset (0.01 .. 1).
            :return: Instance of class NNetEn_entropy
        """

        super().__init__(database=database, mu=mu)
        self.database = database
        conf_layers = [self.norm_db.shape[1] - 1,
                            self.__NEURON_HID_LAYER,
                            self.lab_test.shape[1]]
        self.conf_layers = List()
        [self.conf_layers.append(x) for x in conf_layers]

        self.W1 = np.zeros([self.conf_layers[1],
                            self.conf_layers[0] + 1], dtype=np.float64)
        self.W2 = np.zeros([self.conf_layers[2],
                            self.conf_layers[1] + 1], dtype=np.float64)

    def __validation_input_params(self, time_series : np.ndarray, method : int,
                                        epoch : int, metric : str, log : bool):
        """
        Validation of input data
            :param time_series: type -> np.ndarray
            :param epoch: type -> int [> 0]
            :param method: type -> int [1 .. 6]
            :param metric: type -> str
            :param log: type -> bool
            :return: If the validation of the type and data was
                not successful, an error message will be displayed
        """

        if not (isinstance(log, bool) and
                isinstance(time_series, np.ndarray)):
            logging.error('Invalid input parameter types')
            exit(0)

        if not (isinstance(method, int) and
            ([x for x in range(1,7)].count(method))):
            logging.error('Invalid parameter - method')
            exit(0)

        if not (isinstance(epoch, int) and epoch > 0):
            logging.error('Invalid parameter - ecpoh')
            exit(0)

        if not (isinstance(metric, str) and
                ['Acc', 'R2E', 'PE'].count(metric)):
            logging.error('Invalid parameter - metric')
            exit(0)

    def __preparation_launch_LogNNet(self, time_series: np.ndarray, method: int):
        """
        Preparatory stage.
        Formation of the weight matrix W1 and W2, calculation of values for normalization
        and calculation of neurons of the 1st layer
            :param time_series: input data with a time series in numpy array format
            :param method: One of 6 methods for forming a reservoir matrix from the time series M1 ... M6
        """

        # Step 3
        logging.debug('Load time series, method - %s', method)
        self.W1 = formation_W1_coef_time_series(time_series,
                                                self.W1, method)

        # Step 4
        logging.debug('Calculating min, max and average coefficients')
        self.Sh_max, self.Sh_min, self.Sh_mean = \
                    calculation_normalization_params(
                        self.norm_db, self.W1)

        # Step 5
        logging.debug('Calculation neurons first layer LogNNet.')
        self.__Sh = calculation_neuron_first_layer(self.norm_db,
                                            self.W1, self.Sh_max,
                                            self.Sh_min, self.Sh_mean)
        self.__Sh = np.nan_to_num(self.__Sh)

        self.W2[:] = 0.5

    def __trainng_LogNNet(self, epoch: int):
        """
        Neural network training stage LogNNet
            :param epoch: The number of training epochs for the LogNNet neural network,
                    with a number greater than 0
            :return: Trained weight matrix W2
        """

        # Step 6
        logging.debug('LogNNet %s network training', self.conf_layers)
        start_all = time()
        for ep in range(epoch):
            start_ep = time()
            for index in range(self.norm_db.shape[0]):
                Sout = back_prop_calc_value(self.__Sh[index], self.W2)
                err2 = err_calc_last_layer(self.lab_train[index], Sout)
                self.W2 = weight_training(self.W2, err2,
                                    self.__Sh[index], self.__NORMA)
            logging.debug('Epoch {}. Time - {:.3f} s.'.
                          format(ep, time() - start_ep))

        logging.debug('Time traning network - {:.3f} s.'.
                      format(time() - start_all))

    def __testing_LogNNet(self, metric: str):
        """
        Neural network testing stage LogNNet and entropy calculation
            :param metric: Options: metric = 'Acc', metric = 'R2E', metric = 'PE'
            :return: NNetEn value
        """

        metric_NNetEn = np.zeros(self.lab_test.shape[0])
        # Step 7
        logging.debug('Testing network')
        for index in range(self.norm_db_T.shape[0]):
            Sh_T = back_prop_calc_first_layer(
                self.norm_db_T[index], self.W1,
                self.Sh_max, self.Sh_min, self.Sh_mean)
            Sh_T = np.nan_to_num(Sh_T)
            self.Sout_T = back_prop_calc_value(Sh_T, self.W2)

            metric_NNetEn[index] = calc_metric(
                                    model=self.lab_test[index],
                                    Sout=self.Sout_T,
                                    metric=metric)

        self.NNetEn = np.mean(metric_NNetEn)

    def calculation(self, time_series, method=3,
                        epoch=20, metric='Acc', log=False):
        """
        Command to calculation a NNetEn parameter
            :param time_series: input data with a time series in numpy array format.
            :param epoch: (default = 20). The number of training epochs for the
                        LogNNet neural network, with a number greater than 0.
            :param method: (default = 3) One of 6 methods for forming a reservoir
                        matrix from the time series M1 ... M6.
            :param metric: (default = 'Acc') 'Acc' - accuracy metric,
                        'R2E' - R2 Efficiency metric,
                        'PE' - Pearson Efficiency metric.
            :param log: (default = False) Parameter for logging the main data used
                        in the calculation. Recording is done in log.txt file.
            :return: NNetEn – the entropy value for the given parameters.
        """

        self.__validation_input_params(time_series=time_series,
                                     method=method, epoch=epoch,
                                     metric=metric, log=log)

        time_work = time()
        self.__preparation_launch_LogNNet(time_series=time_series,
                                            method=method)
        self.__trainng_LogNNet(epoch=epoch)
        self.__testing_LogNNet(metric=metric)

        logging.debug('NNetEn = {}'.format(self.NNetEn))
        logging.debug('Time calculation NNetEn - {:.3f} s.'.
                     format(time() - time_work))

        if (log == True):
            logging.debug('Write results to file - log.txt')
            self.__logging(time() - time_work, epoch, time_series.size, metric)

        return self.NNetEn

    def __logging(self, calc_time, epoch, time_serises_size,
                  metric, log_file='log.txt'):
        """
        Logging of the main parameters used during the operation of the module NNetEn
            :param calc_time: Enropy calculation time NNetEn in seconds
            :param epoch: The number of training epochs of the neural network LogNNet
            :param time_serises_size: Time series array size
            :param metric: Metric for calculating entropy value
            :param log_file: Log file name
            :return: The file is created with the parameters used in the calculations
        """
        dist_log = {
            "NNetEn" : str(self.NNetEn),
            "Calculation Time" : round(calc_time, 4),
            "Epoch" : epoch,
            "Metric" : metric,
            "W1 size" : self.W1.size,
            "Database" : self.database,
            "Coef. mu" : self.mu,
            "Length Time Series" : time_serises_size
        }

        isfile = os.path.isfile(log_file)
        with open(log_file, mode='a', encoding='utf-8') as file:
            if not isfile:
                file.write("\t".join(dist_log.keys()) + "\n")

            file.write("\t".join([str(value) for _, value in
                                  dist_log.items()]) + "\n")

def main():
    pass

if __name__ == '__main__':
    main()


from setuptools import setup, find_packages

VERSION = '1.0.4'

def readme():
  with open('README.md', 'r') as f:
    return f.read()

requirements = [
    'numba==0.56.2',
    'numpy==1.23.3'
]

setup(
    name='NNetEn',
    version=VERSION,
    description='Python Package for Neural '
                'Network Entropy (NNetEn) calculation',
    long_description=readme(),
    long_description_content_type="text/markdown",
    url='https://github.com/izotov93/NNetEn',
    author="Yuriy Izotov et al.",
    author_email='izotov93@yandex.ru',
    install_requires=requirements,
    packages=find_packages(),
    classifiers=[
        'Development Status :: 5 - Production/Stable',
		'Programming Language :: Python :: 3.8',
		'Operating System :: OS Independent',
        'License :: OSI Approved :: BSD License',
	            ],
    python_requires='>=3.8',
    package_data={"NNetEn.Database": ["*.txt",
                                      "*.idx3-ubyte",
                                      "*.idx1-ubyte",
                                      "*.VM"]},
    include_package_data=True,
)
